﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;



namespace bvctr {
	public class PenisCollision : MVRScript {
		private List<GameObject> debugObjects = new List<GameObject>();

		/*
		* Scale value to produce estimated visual colliders.
		* Without this, copying rigidbody colliders transform results in objects larger than a default person atom.
		*/
		private float estimatedScale = 0.08f;

		// Rigidbodies
		protected Rigidbody PenisBase;
		protected Rigidbody PenisMid;
		protected Rigidbody PenisTip;
		// protected Rigidbody TestesLeft;
		// protected Rigidbody TestesMain;
		// protected Rigidbody TestesRight;
		protected Rigidbody Testes;

		// UI
		protected JSONStorableBool isBaseEnabled;
		protected JSONStorableBool isDebugEnabled;
		protected JSONStorableBool isMidEnabled;
		protected JSONStorableBool isTestesEnabled;
		protected JSONStorableBool isTipEnabled;
		// protected JSONStorableBool isTestesMainEnabled;
		// protected JSONStorableBool isTestesLeftEnabled;
		// protected JSONStorableBool isTestesRightEnabled;
		protected JSONStorableFloat colliderScale = new JSONStorableFloat("Scale", 1f, 0.01f, 1f, true);



		private void DebugClear(bool restart = false) {
			debugObjects.ForEach((GameObject obj) => {
				Destroy(obj, 0f);
			});

			if (restart) {
				DebugStart();
			}
		}

		/**
		* Add estimated visualization for colliders.
		*/
		private void DebugStart() {
			Dictionary<string, Rigidbody> colliderMapping = new Dictionary<string, Rigidbody>();
			Component[] comps = PenisBase.GetComponentsInChildren(typeof(CapsuleCollider));

			colliderMapping.Add("AutoColliderGen1Hard", this.PenisBase);
			colliderMapping.Add("AutoColliderGen2Hard", this.PenisMid);
			colliderMapping.Add("AutoColliderGen3aHard", this.PenisTip);
			colliderMapping.Add("AutoColliderGen3bHard", this.PenisTip);
			// colliderMapping.Add("AutoColliderTestes", this.TestesMain);

			foreach (Component c in comps) {
				GameObject cap = GameObject.CreatePrimitive(PrimitiveType.Capsule);
				cap.name = $"{c.name}_debug";
				Rigidbody rb = new Rigidbody();
				Renderer renderer = cap.GetComponent<Renderer>();
				renderer.material.shader = Shader.Find("Transparent/Diffuse");
				renderer.material.color = new Color(0f, 0f, 1f, 0.2f);

				// Set color
				if (colliderMapping.ContainsKey(c.name)) {
					if (!colliderMapping[c.name].detectCollisions) {
						renderer.material.color = new Color(1f, 0f, 0f, 0.2f);
					}
				}

				// Remove physics
				rb = cap.AddComponent<Rigidbody>();
				rb.detectCollisions = false;
				rb.useGravity = false;

				this.debugObjects.Add(cap);

				// Set transform
				cap.transform.SetParent(c.transform);
				cap.transform.localScale = c.transform.localScale * estimatedScale;
				cap.transform.localPosition = Vector3.zero;
				cap.transform.rotation = c.transform.rotation;
			}
		}

		/**
		* Transform scale of rigidbody colliders.
		*/
		private void ScaleColliders(float scale) {
			Component[] comps = PenisBase.GetComponentsInChildren(typeof(CapsuleCollider));

			foreach (Component c in comps) {
				CapsuleCollider col = c as CapsuleCollider;
				Vector3 newScale = new Vector3(1f, scale * 0.9f, 1f) * colliderScale.val;

				//if (col.direction == 2) {
				//	newScale = new Vector3(scale, 1f, 1f) * colliderScale.val;
				//}

				c.transform.localScale = newScale;
			}

			if (isDebugEnabled.val) {
				DebugClear(true);

				//isDebugEnabled.valNoCallback = false;
			}
		}

		/**
		* Turn rigidbody collision on or off.
		*/
		private void ToggleCollision(Rigidbody rb, bool state) {
			rb.detectCollisions = state;

			if (isDebugEnabled.val) {
				DebugClear(true);
			}
		}



		public IEnumerator DeferredInit() {
			// yield return new WaitForEndOfFrame();

			int waited = 0;

			while (waited < 6) {
				yield return new WaitForSeconds(1);
				// yield return new WaitForFixedUpdate();

				waited++;
			}

			// Apply current settings
			ToggleCollision(this.PenisBase, this.isBaseEnabled.val);
			ToggleCollision(this.PenisMid, this.isMidEnabled.val);
			ToggleCollision(this.PenisTip, this.isTipEnabled.val);
			ToggleCollision(this.Testes, this.isTestesEnabled.val);

			containingAtom.RestoreFromLast(this);
		}

		public override void Init() {
			try {


/*
string rbs = "";
foreach (Rigidbody rb in containingAtom.rigidbodies) {
	if (
		rb.name.Contains("Female") ||
		rb.name.Contains("pelvis") ||
		rb.name.Contains("Pony") ||
		rb.name.Contains("Shin")
	) {
		continue;
	}

	rbs += $"{rb.name}\n";
}
SuperController.LogMessage(rbs);
*/


				// Check plugin is added to male Person Atom
				if (containingAtom.type != "Person") {
					SuperController.LogMessage("PenisCollision plugin needs to be added to a Person atom.");
					return;
				}

				// Init rigidbodies
				this.PenisBase = containingAtom.rigidbodies.Where((rb) => rb.name == "Gen1").First();
				this.PenisMid = containingAtom.rigidbodies.Where((rb) => rb.name == "Gen2").First();
				this.PenisTip = containingAtom.rigidbodies.Where((rb) => rb.name == "Gen3").First();
				this.Testes = containingAtom.rigidbodies.Where((rb) => rb.name == "Testes").First();

				CreateSpacer().height = 40f;

				this.isBaseEnabled = new JSONStorableBool("PenisBase collision", true);
				this.isBaseEnabled.setCallbackFunction = (bool val) => {
					this.ToggleCollision(this.PenisBase, val);
				};
				RegisterBool(this.isBaseEnabled);
				CreateToggle(this.isBaseEnabled);

				this.isMidEnabled = new JSONStorableBool("PenisMid collision", true);
				this.isMidEnabled.setCallbackFunction = (bool val) => {
					this.ToggleCollision(this.PenisMid, val);
				};
				RegisterBool(this.isMidEnabled);
				CreateToggle(this.isMidEnabled);

				this.isTipEnabled = new JSONStorableBool("PenisTip collision", true);
				this.isTipEnabled.setCallbackFunction = (bool val) => {
					this.ToggleCollision(this.PenisTip, val);
				};
				RegisterBool(this.isTipEnabled);
				CreateToggle(this.isTipEnabled);

				this.isTestesEnabled = new JSONStorableBool("Testes collision", true);
				this.isTestesEnabled.setCallbackFunction = (bool val) => {
					this.ToggleCollision(this.Testes, val);
				};
				RegisterBool(this.isTestesEnabled);
				CreateToggle(this.isTestesEnabled);

				// Debug toggle
				this.isDebugEnabled = new JSONStorableBool("Visualize", false);
				this.isDebugEnabled.setCallbackFunction = (bool val) => {
					if (val) {
						this.DebugStart();
					} else {
						this.DebugClear();
					}
				};
				CreateToggle(this.isDebugEnabled, true);

				this.colliderScale.setCallbackFunction = (float val) => {
					this.ScaleColliders(val);
				};
				RegisterFloat(this.colliderScale);
				CreateSlider(this.colliderScale, true);

				StartCoroutine(DeferredInit());
			} catch (Exception e) {
				SuperController.LogError("PenisCollision.cs Exception: " + e);
			}
		}

		public void OnDestroy() {
			this.DebugClear();

			// Reset collider scales
			this.ScaleColliders(1f);

			// Enable all collision
			this.ToggleCollision(this.PenisBase, true);
			this.ToggleCollision(this.PenisMid, true);
			this.ToggleCollision(this.PenisTip, true);
			this.ToggleCollision(this.Testes, true);
		}
	}
}
